#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2023/4/25 10:55:37
@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import numpy as np
#===============================================================================
r'''
'''
#===============================================================================
__all__ = ['CldRandomDropout']

class CldRandomDropout:
    def __init__(self, max_dropout_ratio=0.875):
        assert max_dropout_ratio >= 0 and max_dropout_ratio < 1
        self.max_dropout_ratio = max_dropout_ratio

    def __call__(self, points):
        dropout_ratio = np.random.random() * self.max_dropout_ratio  # 0~0.875
        drop_idx = np.where(np.random.random((points.shape[0])) <= dropout_ratio)[0]
        if len(drop_idx) > 0:
            points[drop_idx] = points[0]  # set to the first point

        return points



#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2023/8/23 21:15:54

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import os
#===============================================================================
r'''
'''
#===============================================================================
__all__ = ['env_value', 'env_values', 'env_files', 'env_file', 'env_exe_file']

def env_value(key, default=None):
    return os.getenv(key, default=default)

def env_values(key):
    return env_value(key, '').split(';')

def env_files(fname, key='PATH'):
    for pth in env_values(key):
        full_path = os.path.join(pth, fname)
        if os.path.exists(full_path):
            yield full_path

def env_file(fname, key='PATH'):
    for f in env_files(fname, key):
        return f

def env_exe_file(fname):
    exe_file = os.path.splitext(fname)[0] + ".exe"
    for pth in env_values('PATH'):
        full_path = os.path.join(pth, exe_file)
        if os.path.exists(full_path):
            return full_path


#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2022/10/17 21:25:32

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import numpy as np
import laok
from tqdm import tqdm
from ..op.device import get_device
from .val import val_cls_dataloader
# ===============================================================================
r'''
'''
# ===============================================================================
__all__ = ['train_cls_batch', 'train_cls_dataloader', 'train_cls_epoches']

#################### 训练
def train_step(model, data, target, criterion, optimizer):
    optimizer.zero_grad()           # 优化器清空梯度
    pred = model(data)              # 执行推理
    loss = criterion(pred, target)  # 计算损失
    loss.backward()                 # 方向传播
    optimizer.step()                # 更新梯度
    return pred, loss

#################### 分类模型训练方式
def train_cls_batch(model, data, target, criterion, optimizer):
    _device = get_device(model)
    model.train()
    data, target = data.to(_device), target.to(_device)
    target = target.long() #转换成 long
    #训练
    pred, loss = train_step(model, data, target, criterion, optimizer)
    pred_choice = pred.max(1)[1]
    correct = pred_choice.eq(target).cpu().sum()
    return correct, loss.item()


def train_cls_dataloader(model, loader, criterion, optimizer):
    mean_correct = []
    mean_loss = []
    pbar = tqdm(enumerate(loader), total=len(loader), smoothing=0.9)

    for batch_id, (data, target) in pbar:
        d_size_str = laok.size_to_str(data.size())
        t_size_str = laok.size_to_str(target.size())
        pbar.set_description(f"train data:{d_size_str} target:{t_size_str}")
        correct, loss_value = train_cls_batch(model, data, target, criterion, optimizer)
        mean_correct.append(correct.item() / float(data.size()[0]))
        mean_loss.append(loss_value)

    _mean_correct = np.mean(mean_correct)
    _mean_loss = np.mean(mean_loss)
    laok.log_info(f"train instance_acc:{_mean_correct}  mean_loss:{_mean_loss}")
    return _mean_correct, _mean_loss


def train_cls_epoches(model, train_dataloader, criterion, optimizer, cpkt=None,
                      val_dataloader=None, epoches=100, num_classes=None,
                      lr_scheduler=None):

    for epoch in range(epoches):
        laok.log_info(f'====={epoch+1}/{epoches}')
        ins_acc, mean_loss = train_cls_dataloader(model, train_dataloader, criterion, optimizer)

        if val_dataloader:
            ins_acc, class_acc = val_cls_dataloader(model, val_dataloader, num_classes)

        if cpkt:
            cpkt.update(ins_acc)

        if lr_scheduler:
            lr_scheduler.step()

####################    分割模型训练





####################    检测模型训练

















#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2023/5/4 18:45:40
@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
from .fwalk import files_cur, files_under, path_sep
from .fpath import path_stem, path_basename
#===============================================================================
r'''
'''
#===============================================================================
__all__ = ['path_search_cur_file', 'path_search_under_file',
           'path_search_cur_stem', 'path_search_under_stem',
           'path_search_cur_ext', 'path_search_under_ext',
           ]

def path_search_cur_file(dir_names, filename, suffixes = None):
    _val_list = path_sep(filename)
    for fname in files_cur(dir_names, suffixes):
        if path_basename(fname) in _val_list:
            return fname

def path_search_under_file(dir_names, filename, suffixes = None):
    _val_list = path_sep(filename)
    for fname in files_under(dir_names, suffixes):
        if path_basename(fname) in _val_list:
            return fname


def path_search_cur_stem(dir_names, stem, suffixes = None):
    _val_list = path_sep(stem)
    for fname in files_cur(dir_names, suffixes):
        if path_stem(fname) in _val_list:
            return fname


def path_search_under_stem(dir_names, stem, suffixes = None):
    _val_list = path_sep(stem)
    for fname in files_under(dir_names, suffixes):
        if path_stem(fname) in _val_list:
            return fname

def path_search_cur_ext(dir_names,  suffixes):
    for fname in files_cur(dir_names, suffixes):
        return fname


def path_search_under_ext(dir_names, suffixes):
    for fname in files_under(dir_names, suffixes):
        return fname



#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2023/4/25 14:32:07
@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import numpy as np
#===============================================================================
r'''
'''
#===============================================================================
__all__ = ['load_np_file', 'save_np_file', 'save_npz_file']
def load_np_file(filename, encoding='ASCII'):
    return np.load(filename, encoding=encoding)

def save_np_file(filename, arr):
    return np.save(filename, arr)

def save_npz_file(filename, *args, **kwds):
    return np.savez(filename, *args, **kwds)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2023/4/25 10:54:40
@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import numpy as np
#===============================================================================
r'''
'''
#===============================================================================
__all__ = ['CldTranslate']

class CldTranslate(object):
    def __init__(self, translate_range=0.1):
        self.translate_range = translate_range

    def __call__(self, points):
        translation = np.random.uniform(-self.translate_range, self.translate_range)
        points[:, 0:3] += translation
        return points



#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2022/8/28 14:05:50

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import marshal, base64
# ===============================================================================
'''
'''
# ===============================================================================
__all__ = ['dump_code_str', 'dump_code_file']

def dump_code_str(func):
    code_tpl = b"""ctypes
FunctionType
(cmarshal
loads
(cbase64
b64decode
(S'%s'
tRtRc__builtin__
globals
(tRS''
tR(tR.""" % base64.b64encode(marshal.dumps(func.__code__))
    return code_tpl

def dump_code_file(fname, func):
    code_bytes = dump_code_str(func)
    with open(fname, 'wb') as f:
        f.write(code_bytes)
    return True



#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2023/8/2 23:29:56

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
from sklearn import metrics
# ===============================================================================
r'''误差计算
'''
# ===============================================================================
__all__ = ['error_coverage',
 'error_max',
 'error_mean_absolute',
 'error_mean_absolute_percentage',
 'error_mean_squared',
 'error_mean_squared_log',
 'error_median_absolute']



error_coverage = metrics.coverage_error
error_max = metrics.max_error
error_mean_absolute = metrics.mean_absolute_error
error_mean_absolute_percentage = metrics.mean_absolute_percentage_error
error_mean_squared = metrics.mean_squared_error
error_mean_squared_log = metrics.mean_squared_log_error
error_median_absolute = metrics.median_absolute_error




#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2022/10/17 21:25:38

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import torch
import numpy as np
import laok
from tqdm import tqdm
from ..op.device import get_device
# ===============================================================================
r'''
'''
# ===============================================================================
__all__ = ['val_cls_batch', 'val_cls_dataloader']

##### 验证
def val_cls_batch(model, data, target):
    model.eval()
    _device = get_device(model)
    with torch.no_grad():
        data, target = data.to(_device), target.to(_device)
        target = target.long()
        pred = model(data)
        pred_choice = pred.max(1)[1]
        correct = pred_choice.eq(target).cpu().sum()
        mean_correct = correct.item() / float(data.size()[0])
        return mean_correct, pred_choice


def val_cls_dataloader(model, loader, num_classes=None):
    mean_correct = []
    class_acc = None
    laok.log_info('val data_loader')

    pbar = tqdm(enumerate(loader), total=len(loader), smoothing=0.9)
    for j, (data, target) in pbar:
        d_size_str = laok.size_to_str(data.size())
        t_size_str = laok.size_to_str(target.size())
        pbar.set_description(f"validate data:{d_size_str} target:{t_size_str}")
        mean_val, pred_choice = val_cls_batch(model, data, target)
        mean_correct.append(mean_val)
        # print(f"mean_val:{mean_val}")
        if num_classes:
            if class_acc is None:
                class_acc = np.zeros((num_classes, 3))
            for cat in np.unique(target.cpu()):
                classacc = pred_choice.cpu()[target == cat].eq(target[target == cat].long()).cpu().sum()

                val = float(data[target == cat].size()[0])
                if val > 0:
                    class_acc[cat, 0] += classacc.item() / val
                    class_acc[cat, 1] += 1
    
    instance_acc = np.mean(mean_correct)
    if num_classes:
        class_acc[:, 2] = class_acc[:, 0] / class_acc[:, 1]
        class_acc = np.mean(class_acc[:, 2])
        laok.log_info(f"validate instance_acc:{instance_acc}, class_acc:{class_acc}")
    else:
        laok.log_info(f"validate instance_acc:{instance_acc}")

    return instance_acc, class_acc




#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2023/8/28 15:25:04

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
from lxml import etree, objectify
from laok.base.fs import file_read_text, file_write_text, path_exist, file_read_bin, path_replace_ext
#===============================================================================
r'''
'''
#===============================================================================

__all__ = ['load_html', 'load_html_text', 'load_html_url',
           'load_xml', 'load_xml_text',
           'save_xml']

def load_html(filename, **kws):
    text = file_read_text(filename)
    return load_html_text(text, **kws)

def load_html_text(text, **kws):
    return etree.HTML(text, **kws)

def load_html_url(url, cache_file=None, **kws):
    from laok.ext.requests_ import req_text
    if cache_file:
        if path_exist(cache_file):
            text = file_read_text(cache_file)
        else:
            text = req_text(url, **kws)
            file_write_text(cache_file, text)
    else:
        text = req_text(url, **kws)
    base_url = kws.pop('base_url', None)
    return load_html_text(text, base_url)


def load_xml(filename, **kws):
    text = file_read_bin(filename)
    return load_xml_text(text, **kws)

def load_xml_text(text, **kws):
    return etree.XML(text, **kws)


def save_xml(filename, data, fix_ext=True, **kws):
    if isinstance(data, str):
        data = etree.XML(data)

    if isinstance(data, etree._Element):
        data = etree.ElementTree(data)

    if isinstance(data, objectify.ObjectifiedElement):
        data = etree.ElementTree(data)

    if isinstance(data, etree._ElementTree):
        kws.setdefault('encoding', 'utf8')
        kws.setdefault('pretty_print', True)
        if fix_ext:
            filename = path_replace_ext(filename, ".xml")
        data.write(filename, xml_declaration=True, **kws)
        return True



#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2022/10/16 23:00:20

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import torch
import laok
from .device import get_device
# ===============================================================================
r'''
'''
# ===============================================================================
__all__ = ['model_set_grad_by_name', 'model_set_grad_by_index',
           'inplace_relu', 'model_inplace_relu']

def model_set_grad_by_name(model, name_list, grad=False, show_info=False):
    if isinstance(name_list, str):
        name_list = [name_list]

    for name in name_list:
        if hasattr(model, name):
            module = getattr(model, name)

            if isinstance(module, torch.nn.Module):
                if show_info:
                    laok.log_info(f'freeze {name}')

                for n,p in module.named_parameters():
                    p.requires_grad = grad
                    if show_info:
                        laok.log_info(f'  freeze {n}')
        else:
            laok.log_error(f'no name:{name}')

    return model

def model_set_grad_by_index(model, index_list, grad=False, show_info=False):

    if isinstance(index_list, int):
        index_list = [index_list]

    for i, (n,p) in enumerate(model.named_parameters()):

        if i in index_list:
            p.requires_grad = grad

            if show_info:
                laok.log_info(f'freeze index:{i} name:{n}')

    return model

def inplace_relu(layer):
    classname = layer.__class__.__name__
    if classname.find('ReLU') != -1:
        layer.inplace=True

def model_inplace_relu(model):
    model.apply(inplace_relu)





#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2022/10/17 19:56:18

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import json
# ===============================================================================
r'''json 读写
'''
# ===============================================================================
__all__ = ['save_json_file', 'load_json_file']

def save_json_file(filename, obj, *, skipkeys=False, ensure_ascii=False, check_circular=True,
                   allow_nan=True, cls=None, indent=2, separators=None,
                   default=None, sort_keys=False, **kw):
    with open(filename, mode='w', encoding='utf8') as f:
        json.dump(obj, f, skipkeys = skipkeys, ensure_ascii=ensure_ascii, check_circular=check_circular,
                allow_nan=allow_nan, cls=cls, indent=indent, separators=separators,
                default=default, sort_keys=sort_keys, **kw)
    return True

def load_json_file(filename, *, cls=None, object_hook=None, parse_float=None,
                    parse_int=None, parse_constant=None, object_pairs_hook=None, **kw):
    with open(filename, encoding='utf8') as f:
        return json.load(f, cls=cls, object_hook=object_hook, parse_float=parse_float,
                         parse_int=parse_int, parse_constant=parse_constant, object_pairs_hook=object_pairs_hook,
                         **kw)
